#include <QTRSensors.h>

#define NUM_SENSORS 8 // number of sensors used
#define TIMEOUT 4 // waits for 2500 microseconds for sensor outputs to go low
#define EMITTER_PIN 13 // emitter is controlled by digital pin 13

QTRSensors qtra;

const uint8_t SensorCount = 8;
uint16_t sensorValues[SensorCount];

int IndicatorLED = 13;
//Motor 1
int pinAIN1 = 7; //Direction
int pinAIN2 = 6; //Direction
int pinPWMA = 5; //Speed

//Motor 2
int pinBIN1 = 8; //Direction
int pinBIN2 = 9; //Direction
int pinPWMB = 10; //Speed

//Standby
int pinSTBY = 12;

int position = 0;
int error = 0;
int I = 0;

int m1Speed = 0;
int m2Speed = 0;
int motorSpeed = 0;
unsigned long previousMillis = 0; // time keeping function
int BlinkTime = 300; // blink time
int CycleTime = 1; // used to call for a blink toggle
int BlinkCycle = 1; // used to alternate the LED on and off
///////////////////////PID TUNING//////////////
int lastError = 0;
float KP = 25; // experiment to determine this, start by something small that just makes your bot follow the line at a slow speed
float Ki = 0.2;
float KD = 30; // experiment to determine this, slowly increase the speeds and adjust this value. ( Note: Kp < Kd)

int M1 = 30; //base motor speeds
int M2 = 30; //base motor speeds
int M1max = 50; //max motor speeds
int M2max = 50; //max motor speeds
int M1min = 0; //min motor speeds
int M2min = 0; //min motor speeds

//Constants to help remember the parameters
static boolean turnCW = 0; //for motorDrive function
static boolean turnCCW = 1; //for motorDrive function
static boolean motor1 = 0; //for motorDrive, motorStop, motorBrake functions
static boolean motor2 = 1; //for motorDrive, motorStop, motorBrake functions


// The path variable will store the path that the robot has taken.  It
// is stored as an array of characters, each of which represents the
// turn that should be made at one intersection in the sequence:
//  'L' for left
//  'R' for right
//  'S' for straight (going straight through an intersection)
//  'B' for back (U-turn)
// You should check to make sure that the path_length of your
// maze design does not exceed the bounds of the array.
char path[100] = "";
unsigned char path_length = 0; // the length of the path

void setup()
{


  qtra.setTypeAnalog();
  qtra.setSensorPins((const uint8_t[]) {
    A2, A4, A6, A7, A1, A0, A5, A3
  }, SensorCount);
  qtra.setEmitterPin(13);

  //Set the PIN Modes
  pinMode(IndicatorLED, OUTPUT);
  digitalWrite(IndicatorLED, HIGH);
  pinMode(pinPWMA, OUTPUT);
  pinMode(pinAIN1, OUTPUT);
  pinMode(pinAIN2, OUTPUT);
  pinMode(pinPWMB, OUTPUT);
  pinMode(pinBIN1, OUTPUT);
  pinMode(pinBIN2, OUTPUT);
  pinMode(pinSTBY, OUTPUT);


  // analogRead() takes about 0.1 ms on an AVR.
  // 0.1 ms per sensor * 4 samples per sensor read (default) * 6 sensors
  // * 10 reads per calibrate() call = ~24 ms per calibrate() call.
  // Call calibrate() 400 times to make calibration take about 10 seconds.

  for (int i = 0; i < 100; i++) // make the calibration take about 10 seconds
  {
    TimeCheck();
    delay(5);

    // auto calibration sweeping left/right, tune 'calSpeed' motor speed at declaration
    if (i == 0 || i == 60) // slow sweeping turn right to pass sensors over line
    {
      digitalWrite(pinAIN1, HIGH);
      analogWrite(pinPWMA, 20);
      digitalWrite(pinBIN2, HIGH);
      analogWrite(pinPWMB, 20);
      digitalWrite(pinAIN2, LOW);
      digitalWrite(pinBIN1, LOW);
    }

    else if (i == 20 || i == 100) // slow sweeping turn left to pass sensors over line
    {
      digitalWrite(pinAIN2, HIGH);
      analogWrite(pinPWMA, 20);
      digitalWrite(pinBIN1, HIGH);
      analogWrite(pinPWMB, 20);
      digitalWrite(pinAIN1, LOW);
      digitalWrite(pinBIN2, LOW);
    }
    qtra.calibrate(); // reads all sensors 10 times at 2500 us per read (i.e. ~25 ms per call)
  }

  // read calibrated sensor values and obtain a measure of the line position from 0 to 7000
  // To get raw sensor values, call:
  //  qtrrc.read(sensorValues); instead of unsigned int position = qtra.readLineBlack(sensorValues);
  position = qtra.readLineBlack(sensorValues);

  while (sensorValues[6] < 200)  // wait for line position to near center
  {
    position = qtra.readLineBlack(sensorValues);
  }

  // slow down speed
  analogWrite(pinPWMA, 30);
  analogWrite(pinPWMB, 30);

  // find center
  while (position > 4350)  // wait for line position to find center
  {
    position = qtra.readLineBlack(sensorValues);
  }

  // stop both motors
  analogWrite(pinPWMB, 0);  // stop right motor first which kinda helps avoid over run
  analogWrite(pinPWMA, 0);

} // end setup

void loop()
{
  TimeCheck(); //Check to see if it is time to blink the LED
  PID ();
}

void PID ()
{
  lastError = 0;

  while (1)
  {

    uint16_t position = qtra.readLineBlack(sensorValues);

    switch (position)
    {

      //Line has moved off the left edge of sensor
      case 0:
        digitalWrite(pinAIN1, HIGH);
        analogWrite(pinPWMA, 50);
        digitalWrite(pinBIN2, HIGH);
        analogWrite(pinPWMB, 50);
        digitalWrite(pinAIN2, LOW);
        digitalWrite(pinBIN1, LOW);
        Serial.println("Rotate Left\n");
        break;

      // Line had moved off the right edge of sensor
      case 7000:
        digitalWrite(pinAIN2, HIGH);
        analogWrite(pinPWMA, 50);
        digitalWrite(pinBIN1, HIGH);
        analogWrite(pinPWMB, 50);
        digitalWrite(pinAIN1, LOW);
        digitalWrite(pinBIN2, LOW);

        Serial.println("Rotate Right\n");
        break;

      // compute our "error" from the line position
      // error is zero when the middle sensor is over the line,
      default:
        error = position - 3500 ; /// using an 8 sensor array targeting middle of position 4 to 5
        I = I + error;
        // set the motor speed based on proportional and derivative PID terms
        // kp is the a floating-point proportional constant (maybe start with a value around 0.5)
        // kd is the floating-point derivative constant (maybe start with a value around 1)
        // note that when doing PID, it's very important you get your signs right, or else the
        // control loop will be unstable

        // set the motor speed based on proportional and derivative PID terms
        motorSpeed = KP * error + I * Ki + KD * (error - lastError) ;
        lastError = error;

        //this codes limits the motorSpeed (motor speed pwm value)
        // limit motorSpeed to 55
        if (motorSpeed > 55)
        {

          motorSpeed = 55;
        }

        if (motorSpeed < -55)
        {
          motorSpeed = -55;
        }


        // set motor speeds using the two motor speed variables above 255

        /////////CHANGE THE - OR + BELOW DEPENDING ON HOW YOUR DRIVER WORKS
        m1Speed = M1 - motorSpeed; // M1 and M2 are base motor speeds.
        m2Speed = M2 + motorSpeed;

        if (m1Speed < M1min) //keep speeds to 0 or above
        {
          m1Speed = M1min;
        }
        if (m2Speed < M2min)
        {
          m2Speed = M2min;
        }
        if (m1Speed > M1max) //maximum allowed value
        {
          m1Speed = M1max;
        }
        if (m2Speed > M2max) //maximum allowed value
        {
          m2Speed = M2max;
        }

        forward(); // run the motors based on error information
    }

    if (sensorValues[1] < 100 && sensorValues[2] < 100 && sensorValues[3] < 100 && sensorValues[4] < 100 && sensorValues[5] < 100 && sensorValues[6] < 100)
    {
      // There is no line visible ahead, and we didn't see any
      // intersection.  Must be a dead end.

      /*  digitalWrite(pinAIN1, LOW);
        digitalWrite(pinAIN2, LOW); //This is the opposite of the AIN1
        analogWrite(pinPWMA, motorSpeed);
        digitalWrite(pinBIN1, LOW);
        digitalWrite(pinBIN2, LOW); //This is the opposite of the BIN1
        analogWrite(pinPWMB, motorSpeed); */
      return;
    }

    else if (sensorValues[0] > 200 || sensorValues[7] > 200)
    {
      // Found an intersection.
      return;
    }
  }
}


//////////////////////FORWARD////////
//Drive both motors CW, full speed
void forward()
{
  TimeCheck();
  motorDrive(motor1, turnCW, m1Speed);
  motorDrive(motor2, turnCW, m2Speed);
}

// Turns to the sent variable of
// 'L' (left), 'R' (right), 'S' (straight), or 'B' (back)
// Tune '30' at declaration
void turn(char dir)
{
  switch (dir)
  {
    // Turn left 90deg
    case 'L':
      digitalWrite(pinAIN1, HIGH);
      analogWrite(pinPWMA, 30);
      digitalWrite(pinBIN2, HIGH);
      analogWrite(pinPWMB, 30);
      digitalWrite(pinAIN2, LOW);
      digitalWrite(pinBIN1, LOW);

      position = qtra.readLineBlack(sensorValues);

      while (sensorValues[6] < 200) // wait for outer most sensor to find the line
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // slow down speed
      analogWrite(pinPWMA, 30);
      analogWrite(pinPWMB, 30);

      // find center
      while (position > 4350)  // tune - wait for line position to find near center
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // stop both motors
      analogWrite(pinPWMB, 0);  // stop right motor first to better avoid over run
      analogWrite(pinPWMA, 0);
      break;

    // Turn right 90deg
    case 'R':
      digitalWrite(pinAIN2, HIGH);
      analogWrite(pinPWMA, 30);
      digitalWrite(pinBIN1, HIGH);
      analogWrite(pinPWMB, 30);
      digitalWrite(pinAIN1, LOW);
      digitalWrite(pinBIN2, LOW);

      position = qtra.readLineBlack(sensorValues);

      while (sensorValues[1] < 200) // wait for outer most sensor to find the line
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // slow down speed
      analogWrite(pinPWMA, 30);
      analogWrite(pinPWMB, 30);

      // find center
      while (position < 3250)  // tune - wait for line position to find near center
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // stop both motors
      analogWrite(pinPWMA, 0);
      analogWrite(pinPWMB, 0);
      break;

    // Turn right 180deg to go back
    case 'B':
      digitalWrite(pinAIN2, HIGH);
      analogWrite(pinPWMA, 30);
      digitalWrite(pinBIN1, HIGH);
      analogWrite(pinPWMB, 30);
      digitalWrite(pinAIN1, LOW);
      digitalWrite(pinBIN2, LOW);

      position = qtra.readLineBlack(sensorValues);

      while (sensorValues[1] < 200) // wait for outer most sensor to find the line
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // slow down speed
      analogWrite(pinPWMA, 30);
      analogWrite(pinPWMB, 30);

      // find center
      while (position < 3250)  // tune - wait for line position to find near center
      {
        position = qtra.readLineBlack(sensorValues);
      }

      // stop both motors
      analogWrite(pinPWMA, 0);
      analogWrite(pinPWMB, 0);
      break;

    // Straight ahead
    case 'S':
      // do nothing
      break;
  }
} // end turn

// This function decides which way to turn during the learning phase of
// maze solving.  It uses the variables found_left, found_straight, and
// found_right, which indicate whether there is an exit in each of the
// three directions, applying the "left hand on the wall" strategy.

char select_turn(unsigned char found_left, unsigned char found_straight, unsigned char found_right)
{
  // Make a decision about how to turn.  The following code
  // implements a left-hand-on-the-wall strategy, where we always
  // turn as far to the left as possible.
  if (found_left)
    return 'L';
  else if (found_straight)
    return 'S';
  else if (found_right)
    return 'R';
  else
    return 'B';
} // end select_turn

// Path simplification.  The strategy is that whenever we encounter a
// sequence xBx, we can simplify it by cutting out the dead end.  For
// example, LBL -> S, because a single S bypasses the dead end
// represented by LBL.
void simplify_path()
{
  // only simplify the path if the second-to-last turn was a 'B'
  if (path_length < 3 || path[path_length - 2] != 'B')
    return;

  int total_angle = 0;
  int i;
  for (i = 1; i <= 3; i++)
  {
    switch (path[path_length - i])
    {
      case 'R':
        total_angle += 90;
        break;
      case 'L':
        total_angle += 270;
        break;
      case 'B':
        total_angle += 180;
        break;
    }
  }

  // Get the angle as a number between 0 and 360 degrees.
  total_angle = total_angle % 360;

  // Replace all of those turns with a single one.
  switch (total_angle)
  {
    case 0:
      path[path_length - 3] = 'S';
      break;
    case 90:
      path[path_length - 3] = 'R';
      break;
    case 180:
      path[path_length - 3] = 'B';
      break;
    case 270:
      path[path_length - 3] = 'L';
      break;
  }

  // The path is now two steps shorter.
  path_length -= 2;

} // end simplify_path


// This function is called once, from the main loop
void MazeSolve()
{
  // Loop until we have solved the maze.
  while (1)
  {
    // FIRST MAIN LOOP BODY
    PID();

    // Drive straight a bit.  This helps us in case we entered the
    // intersection at an angle.
    digitalWrite(pinAIN2, HIGH);
    analogWrite(pinPWMA, 200);
    digitalWrite(pinBIN2, HIGH);
    analogWrite(pinPWMB, 200);
    digitalWrite(pinAIN1, LOW);
    digitalWrite(pinBIN1, LOW);
    delay(25);

    // These variables record whether the robot has seen a line to the
    // left, straight ahead, and right, whil examining the current
    // intersection.
    unsigned char found_left = 0;
    unsigned char found_straight = 0;
    unsigned char found_right = 0;

    // Now read the sensors and check the intersection type.
    position = qtra.readLineBlack(sensorValues);

    // Check for left and right exits.
    if (sensorValues[0] > 200)
      found_right = 1;
    if (sensorValues[7] > 200)
      found_left = 1;

    // Drive straight a bit more - this is enough to line up our
    // wheels with the intersection.
    digitalWrite(pinAIN2, HIGH);
    analogWrite(pinPWMA, 50);
    digitalWrite(pinBIN2, HIGH);
    analogWrite(pinPWMB, 50);
    digitalWrite(pinAIN1, LOW);
    digitalWrite(pinBIN2, LOW);
    delay(50);

    position = qtra.readLineBlack(sensorValues);
    if (sensorValues[1] > 200 || sensorValues[2] > 200 || sensorValues[3] > 200 || sensorValues[4] > 200 || sensorValues[5] > 200 || sensorValues[6] > 200)
    {
      found_straight = 1;
    }

    // Check for the ending spot.
    // If all six middle sensors are on dark black, we have
    // solved the maze.
    if (sensorValues[1] > 600 && sensorValues[2] > 600 && sensorValues[3] > 600 && sensorValues[4] > 600 && sensorValues[5] > 600 && sensorValues[6] > 600)
    {
      break;
    }

    // Intersection identification is complete.
    // If the maze has been solved, we can follow the existing
    // path.  Otherwise, we need to learn the solution.
    unsigned char dir = select_turn(found_left, found_straight, found_right);

    // Make the turn indicated by the path.
    turn(dir);

    // Store the intersection in the path variable.
    path[path_length] = dir;
    path_length ++;

    // Simplify the learned path.
    simplify_path();
  }
  // Solved the maze!
  // Now enter an infinite loop - we can re-run the maze as many
  // times as we want to.
  while (1)
  {
    //  we just have it waiting patiently to be placed back on the starting line.
    analogWrite(pinPWMA, 0);  // stop both motors
    analogWrite(pinPWMB, 0);
    digitalWrite(pinAIN1, LOW);
    digitalWrite(pinAIN2, LOW);
    digitalWrite(pinBIN1, LOW);
    digitalWrite(pinBIN2, LOW);

    // hold motors while robot is sitting on end point
    position = qtra.readLineBlack(sensorValues);
    while (sensorValues[1] > 200 && sensorValues[2] > 200 && sensorValues[3] > 200 && sensorValues[4] > 200 && sensorValues[5] > 200 && sensorValues[6] > 200)
    {
      position = qtra.readLineBlack(sensorValues);
      delay(50);
    }

    // hold until the start line is seen again when the robot has been placed there again
    while (1)
    {
      position = qtra.readLineBlack(sensorValues);
      if (sensorValues[2] > 200 || sensorValues[3] > 200 || sensorValues[4] > 200 || sensorValues[5] > 200 && sensorValues[0] < 200 && sensorValues[1] < 200 && sensorValues[6] < 200 && sensorValues[7] < 200)
        break;
      delay(50);
    }

    // delay to give you time to let go of the robot
    delay(2000);

    // Re-run the now solved maze.  It's not necessary to identify the
    // intersections, so this loop is really simple.
    int i;
    for (i = 0; i < path_length; i++)
    {
      // SECOND MAIN LOOP BODY
      PID();

      // drive past intersection slightly slower and timed delay to align wheels on line
      digitalWrite(pinAIN2, HIGH);
      analogWrite(pinPWMA, 50);
      digitalWrite(pinBIN2, HIGH);
      analogWrite(pinPWMB, 50);
      digitalWrite(pinAIN1, LOW);
      digitalWrite(pinBIN1, LOW);
      delay(50); // tune time to allow wheels to position for correct turning

      // Make a turn according to the instruction stored in
      // path[i].
      turn(path[i]);
    }

    // Follow the last segment up to the finish.
    PID();

    // drive past intersection slightly slower and timed delay to align wheels on line
    digitalWrite(pinAIN2, HIGH);
    analogWrite(pinPWMA, 50);
    digitalWrite(pinBIN2, HIGH);
    analogWrite(pinPWMB, 50);
    digitalWrite(pinAIN1, LOW);
    digitalWrite(pinBIN1, LOW);
    delay(50); // tune time to allow wheels to position for correct turning

    // Now we should be at the finish!  Now move the robot again and it will re-run this loop with the solution again.

  } // end running solved

} // end MazeSolve

/* //Apply Brakes, then into Standby
  motorBrake(motor1);
  motorBrake(motor2);
  motorsStandby(); */

void TimeCheck() {
  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis > BlinkTime) {
    // save the last time you blinked the LED
    previousMillis = currentMillis;

    BlinkIt();
  }
}

void motorDrive(boolean motorNumber, boolean motorDirection, int motorSpeed)
{
  /*
    This Drives a specified motor, in a specific direction, at a specified speed:
    - motorNumber: motor1 or motor2 ---> Motor 1 or Motor 2
    - motorDirection: turnCW or turnCCW ---> clockwise or counter-clockwise
    - motorSpeed: 0 to 255 ---> 0 = stop / 255 = fast
  */

  boolean pinIn1; //Relates to AIN1 or BIN1 (depending on the motor number specified)

  //Specify the Direction to turn the motor
  //Clockwise: AIN1/BIN1 = HIGH and AIN2/BIN2 = LOW
  //Counter-Clockwise: AIN1/BIN1 = LOW and AIN2/BIN2 = HIGH
  if (motorDirection == turnCW)
  {
    pinIn1 = HIGH;
  }
  else
  {
    pinIn1 = LOW;
  }

  //Select the motor to turn, and set the direction and the speed
  if (motorNumber == motor1)
  {
    digitalWrite(pinAIN1, pinIn1);
    digitalWrite(pinAIN2, !pinIn1); //This is the opposite of the AIN1
    analogWrite(pinPWMA, motorSpeed);
  }
  else
  {
    digitalWrite(pinBIN1, pinIn1);
    digitalWrite(pinBIN2, !pinIn1); //This is the opposite of the BIN1
    analogWrite(pinPWMB, motorSpeed);
  }

  //Finally , make sure STBY is disabled - pull it HIGH
  digitalWrite(pinSTBY, HIGH);
}

void motorBrake(boolean motorNumber)
{
  /*
    This "Short Brake"s the specified motor, by setting speed to zero
  */

  if (motorNumber == motor1) {
    analogWrite(pinPWMA, 0);
  }
  else
  {
    analogWrite(pinPWMB, 0);
  }
}

void motorStop(boolean motorNumber)
{
  /*
    This stops the specified motor by setting both IN pins to LOW
  */
  if (motorNumber == motor1) {
    digitalWrite(pinAIN1, LOW);
    digitalWrite(pinAIN2, LOW);
  }
  else
  {
    digitalWrite(pinBIN1, LOW);
    digitalWrite(pinBIN2, LOW);
  }
}

void BlinkIt() {
  if (BlinkCycle == 1) {
    //Led On
    digitalWrite(IndicatorLED, HIGH);
  }
  if (BlinkCycle == 2) {
    //Led Off
    digitalWrite(IndicatorLED, LOW);
    BlinkCycle = 0; // reset BlinkCycle
  }
  BlinkCycle++; // Set BlinkCycle to second mode
}
